/*
 * Wacom W8001 Startup Utility for Android
 *
 * Copyright (c) 2011 Mike Bagdanoff, Wacom. <mikeb@wacom.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for
 * more details.
 *
 * 
 */

#include <linux/serio.h>
#include <linux/serial.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>

#include "serio-ids.h"

#define CAL_DATA_SIZE 24
#define USER_CAL_DATA_MAX_SIZE 20
#define USER_CAL_DATA_MIN_SIZE 3

#define NORMAL_MODE_QRERY_LENGTH	11
#define NORMAL_MODE_QRERY_HEADER	0xC0

static int L_emmc = -1;

struct sRect {
	int offsetX;
	int offsetY;
	int xy_mix;
	int yx_mix;
	int extentX;
	int extentY;
	int coordinate_cal_flag;
	int on_level;
	int pressure_cal_flag;
};
struct sRect *cal_data =  NULL;

static void cal_data_finalize(void)
{
	if ((int)cal_data != -1 && cal_data != NULL) {
		munmap((void*)cal_data, 0);
	}
	if (L_emmc != -1) {
		fsync(L_emmc);
		close(L_emmc);
	}

	cal_data = NULL;
}

static int cal_data_initialize(void)
{
	static	const	char*	emmc[] = {
		"/dev/block/mmcblk2",
		"/dev/mmcblk2",
	};
	static	char	keyword[] = "Info";

	uint		index;
	FILE*		table;
	char		line[80];
	off_t		offset = 0;
	size_t		length = 0;

	/* Infö̃AhX擾 */
	table = fopen("/sys/module/rawdatatable/parameters/rawdata_param", "r");
	if (table == NULL) {
		/* s */
		return -1;
	}
	while (fgets(line, sizeof line, table) != NULL) {
		const	char*	separator;

		/* "Info :0x????:0x????" T */
		if (strncmp(line, keyword, sizeof keyword - 1) != 0) {
			/* ̍s */
			continue;
		}

		/* AhX\LT */
		separator = strchr(line + sizeof keyword, ':');
		if (separator == NULL) {
			/* Ȃ */
			break;
		}
		offset = strtoul(++separator, NULL, 0);
		/* TCY\LT */
		separator = strchr(separator, ':');
		if (separator == NULL) {
			/* Ȃ */
			break;
		}
		length = strtoul(++separator, NULL, 0);
		break;
	}
	fclose(table);
	if (offset == 0 || length == 0) {
		/* Infö񂪌Ȃ */
		return -1;;
	}

	/* eMMCfoCXJ */
	for (index = 0; index < sizeof emmc / sizeof emmc[0]; index++) {
		L_emmc = open(emmc[index], O_RDWR);
		if (L_emmc != -1) {
			break;
		}
	}
	if (L_emmc == -1) {
		/* s */
		return -1;;
	}

	/* Infö֐ڑ */
	cal_data = (struct sRect *)mmap(NULL, sizeof *cal_data, PROT_READ | PROT_WRITE, MAP_SHARED, L_emmc, offset + 0x4000);
	if ((int)cal_data == -1) {
		/* s */
		close(L_emmc);
		return -1;
	}
	
	return 0;
}

static int readchar(int fd, unsigned char *c, int timeout)
{
	struct timeval tv;
	fd_set set;

	tv.tv_sec = 0;
	tv.tv_usec = timeout * 1000;

	FD_ZERO(&set);
	FD_SET(fd, &set);

	if (!select(fd + 1, &set, NULL, NULL, &tv))
		return -1;

	if (read(fd, c, 1) != 1)
		return -1;

	return 0;
}

static void setline(int fd, int flags, int speed)
{
	struct termios t;

	tcgetattr(fd, &t);

	t.c_cflag = flags | CREAD | HUPCL | CLOCAL;
	t.c_iflag = IGNBRK | IGNPAR;
	t.c_oflag = 0;
	t.c_lflag = 0;
	t.c_cc[VMIN ] = 1;
	t.c_cc[VTIME] = 0;

	cfsetispeed(&t, speed);
	cfsetospeed(&t, speed);

	tcsetattr(fd, TCSANOW, &t);
}

int flash_query_normal(int fd)
{
	int rv, len;
	unsigned char buf[NORMAL_MODE_QRERY_LENGTH];
	unsigned char q_cmd;
	double tick, nTick;
	struct timeval tv;
	
	q_cmd = '0'; // Stop command
	rv = write(fd, &q_cmd, sizeof(q_cmd));
	if (rv <= 0)
		goto err;
	
	usleep(100*1000);
	
	q_cmd = '*'; //normal mode query command
	rv = write(fd, &q_cmd, sizeof(q_cmd));		/* Send querycommand */
	if (rv <= 0)
		goto err;
	
	gettimeofday(&tv, NULL);
	tick = tv.tv_sec + (double)tv.tv_usec*1e-6;
	while ( (read(fd, buf, 1) != 1 ) || (buf[0] != NORMAL_MODE_QRERY_HEADER)){
		gettimeofday(&tv, NULL);
		nTick = (tv.tv_sec + (double)tv.tv_usec*1e-6)-tick;
		if (nTick > 1.0){ /* Time out in 1 seconds */
			printf("Can't read NORMAL_MODE_QRERY_HEADER\n");
			goto err;
		}
	}
	
	len = 1;
	
	gettimeofday(&tv, NULL);
	tick = tv.tv_sec + (double)tv.tv_usec*1e-6;
	do{
		rv = read(fd, buf, 1 );
		if(rv == 1){
			len++;
		}
		gettimeofday(&tv, NULL);
		nTick = (tv.tv_sec + (double)tv.tv_usec*1e-6)-tick;
		if (nTick > 1.0){ /* Time out in 1 seconds */
			printf("Can't read the remaining query responce (%d)\n",len);
			goto err;
		}
	}while(len != NORMAL_MODE_QRERY_LENGTH);
	
	q_cmd = '1'; // 200Hz Scan Mode
	rv = write(fd, &q_cmd, sizeof(q_cmd));
	if (rv <= 0)
		goto err;
	
	usleep(100*1000);

	return 0;
	
err:
	printf("flash_query_normal fail!!\n");
	return -1;
}

int reset_wacom_module(void)
{
	int fd_reset,rv;
	
	//char reset_state;
	fd_reset = open("/sys/kernel/wacom/reset_state",O_WRONLY);
	if (fd_reset < 0 ) {
		printf("Can't open /sys/kernel/wacom/reset_state\n");
		return -1;
	}
	
	rv = write(fd_reset,"0",strlen("0"));
	if ( rv < 0 ) {
		printf("Cant set reset to high %d\n",errno);
		return -1;
	}
	
	usleep(100*1000);

	rv = write(fd_reset,"1",strlen("1"));
	if ( rv < 0 ) {
		printf("Cant set reset to low %d\n",errno);
		return -1;
	}

	usleep(500*1000);

	return 0;
}

struct dev_details {
	int speed;
	int flags;
	unsigned long type;
	int flush;
};

int main(int argc, char **argv)
{
	unsigned long devt;
	int ldisc, ret = 0;
	struct dev_details details = {B38400, CS8, SERIO_W8001,	0 };
	struct serial_struct serial;
	const char *device = NULL;
	int fd, fd_cal, fd_offset_type, fd_user_offset_x, fd_user_offset_y;
	FILE*	file_user_hand;
	FILE*	file_user_offset;
	int i, read_size;
	unsigned char c, user_hand, user_cal[USER_CAL_DATA_MAX_SIZE];
	int retval;

	retval = mknod("/dev/w8001", S_IFCHR | S_IRUSR | S_IWUSR, makedev(66, 0));
	if (retval < 0){
		printf("wac_startup: cannot create /dev/w8001 \n");
	}

	if(argc > 1)
	{
		device = argv[1];
		fd = open(device, O_RDWR | O_NOCTTY | O_NONBLOCK);
		if (fd < 0) 
		{
			fprintf(stderr, "wac_startup: '%s' - %s\n",
				device, strerror(errno));
			return 1;
		}
		
		retval = ioctl(fd, TIOCGSERIAL, &serial);
		if (retval < 0)
			fprintf(stderr, "Cannot obtain serial information \n");
	
		serial.flags |= ASYNC_LOW_LATENCY;
	
		retval = ioctl(fd, TIOCSSERIAL, &serial);
		if (retval < 0)
			fprintf(stderr, "Cannot set low latency mode\n use default latency instead \n");
	
		setline(fd, details.flags, details.speed);
		
		if(flash_query_normal(fd)){
			details.speed = B19200;
			printf("Changed Baud rate to B19200.\n");
			setline(fd, details.flags, details.speed);
			if(reset_wacom_module()){
				close(fd);
				return EXIT_FAILURE;
			}
			if(flash_query_normal(fd)){
				details.speed = B38400;
				printf("Changed Baud rate to B38400.\n");
				setline(fd, details.flags, details.speed);
				if(reset_wacom_module()){
					close(fd);
					return EXIT_FAILURE;
				}
			}
		}
		
		if (details.flush)
			while (!readchar(fd, &c, 100))
				/* empty */;

		ldisc = N_MOUSE;
		if (ioctl(fd, TIOCSETD, &ldisc)) {
			fprintf(stderr, "wac_startup: can't set line discipline\n");
			close(fd);
			return EXIT_FAILURE;
		}

		devt = details.type;

		if (ioctl(fd, SPIOCSTYPE, &devt)) {
			fprintf(stderr, "wac_startup: can't set device type\n");
			close(fd);
			return EXIT_FAILURE;
		}

		retval = EXIT_SUCCESS;

		fd_cal = open("/dev/w8001", O_RDWR);
		retval = cal_data_initialize();
		if (retval < 0){
			printf("Can't get the calibraton data.\n");
		}else if (fd_cal != -1) {
			if (cal_data) {
				if(cal_data->coordinate_cal_flag == 1){
					retval = write(fd_cal, cal_data, CAL_DATA_SIZE);
					if(retval == CAL_DATA_SIZE){
						printf("wac_startup: calibration 24 byte data written \n");
					}else{
						printf("Can't write calibration data\n");
					}
				}else{
					printf("Can't read calibration data\n");
				}
				if(cal_data->pressure_cal_flag == 1){
					write(fd_cal, &cal_data->on_level, sizeof cal_data->on_level);
					if(retval != (sizeof cal_data->on_level)){
						printf("Can't write the pressure calibration data\n");
					}
				}else{
					printf("Can't read the pressure calibration data.\n");
				}
			}
			else {
				printf("wac_startup: cannot find calibration data \n");
			}
			close(fd_cal);
			cal_data_finalize();
		}
		
		//Set user hand setting
		fd_offset_type = open("/sys/kernel/wacom/user_hand", O_RDWR);
		if(fd_offset_type != -1){
			file_user_hand = fopen("/data/system/wacom/wacom_cal_param", "r");
			if(file_user_hand != NULL){
				if(fread(&user_hand, 1, 1, file_user_hand) == 1){
					write(fd_offset_type, &user_hand, 1);
				}
				fclose(file_user_hand);
			}
			close(fd_offset_type);
		}
		
		// Set user calibration setting
		fd_user_offset_x = open("/sys/kernel/wacom/user_offset_x", O_RDWR);
		fd_user_offset_y = open("/sys/kernel/wacom/user_offset_y", O_RDWR);
		if((fd_user_offset_x != -1) && (fd_user_offset_y != -1)){
			file_user_offset = fopen("/data/system/wacom/wacom_cal_offsets", "r");
			if(file_user_offset != NULL){
				read_size = fread(&user_cal, 1, USER_CAL_DATA_MAX_SIZE - 1, file_user_offset);
				if(read_size >= USER_CAL_DATA_MIN_SIZE){
					user_cal[read_size] = '\0';
					for(i = 0; (user_cal[i] != ',') && (user_cal[i] != '\0'); i++);
					write(fd_user_offset_x, user_cal, i);
					if(i+1 < read_size){
						write(fd_user_offset_y, user_cal + i + 1, read_size - i - 1);
					}
				}
				fclose(file_user_offset);
			}
			close(fd_user_offset_x);
			close(fd_user_offset_y);
		}
		
		while(ret >= 0){
			ret = read(fd, NULL, 0);
		}

		ldisc = 0;
		ioctl(fd, TIOCSETD, &ldisc);
		close(fd);
	}

	return retval;
}

